import urllib.parse
from typing import Any

import requests

from beamlit.client import AuthenticatedClient
from beamlit.common import HTTPError, get_settings


class RunClient:
    def __init__(self, client: AuthenticatedClient):
        self.client = client

    def run(
        self,
        resource_type: str,
        resource_name: str,
        environment: str,
        method: str,
        path: str = "",
        headers: dict[str, str] | None = None,
        json: dict[str, Any] | None = None,
        data: str | None = None,
        params: dict[str, str] | None = None,
    ) -> requests.Response:
        settings = get_settings()
        headers = headers or {}
        params = params or {}

        # Build the path
        if path:
            path = f"{settings.workspace}/{resource_type}s/{resource_name}/{path}"
        else:
            path = f"{settings.workspace}/{resource_type}s/{resource_name}"

        client = self.client.get_httpx_client()
        url = urllib.parse.urljoin(settings.run_url, path)

        kwargs = {
            "headers": headers,
            "params": {"environment": environment, **params},
        }
        if data:
            kwargs["data"] = data
        if json:
            kwargs["json"] = json

        response = client.request(method, url, **kwargs)
        if response.status_code >= 400:
            raise HTTPError(response.status_code, response.text)
        return response



"""Decorators for creating function tools with Beamlit and LangChain integration."""

import json
from collections.abc import Callable
from logging import getLogger

from langchain_core.tools import create_schema_from_function

from beamlit.authentication import new_client
from beamlit.common.settings import get_settings
from beamlit.models import Function, FunctionKit
from beamlit.run import RunClient

logger = getLogger(__name__)


def get_remote_function(func: Callable, function: Function):
    settings = get_settings()
    name = (function and function.metadata and function.metadata.name) or func.__name__

    def _partial(*args, **kwargs):
        # Get function signature parameters
        try:
            client = new_client()
            run_client = RunClient(client)
            logger.debug(
                f"Calling remote function: NAME={name}"
                f" PARAMS={kwargs} ENVIRONMENT={settings.environment}"
            )
            response = run_client.run(
                resource_type="function",
                resource_name=name,
                environment=settings.environment,
                method="POST",
                headers={"Content-Type": "application/json"},
                data=json.dumps(kwargs),
            )
            content = response.text
            if response.status_code >= 400:
                content = f"{response.status_code}:{response.text}"
                logger.error(f"Error calling remote function: {content}")
                return f"Error calling remote function: {content}"
            logger.debug(
                f"Response from remote function: NAME={name}"
                f" RESPONSE={content} ENVIRONMENT={settings.environment}"
            )
            if response.headers.get("content-type") == "application/json":
                return response.json()
            return content
        except Exception as e:
            logger.error(f"Error calling function {name}: {e}")
            raise e

    remote_func = _partial
    remote_func.__name__ = func.__name__
    remote_func.__doc__ = func.__doc__
    return remote_func


def kit(bl_kit: FunctionKit = None, **kwargs: dict) -> Callable:
    """Create function tools with Beamlit and LangChain integration."""

    def wrapper(func: Callable) -> Callable:
        if bl_kit and not func.__doc__ and bl_kit.description:
            func.__doc__ = bl_kit.description
        return func

    return wrapper


def function(*args, function: Function = None, kit=False, **kwargs: dict) -> Callable:
    """Create function tools with Beamlit and LangChain integration."""
    settings = get_settings()

    def wrapper(func: Callable) -> Callable:
        if function and not func.__doc__ and function.spec and function.spec.description:
            func.__doc__ = function.spec.description
        if settings.remote:
            remote_func = get_remote_function(func, function)
            if not kwargs.get("args_schema"):
                kwargs["args_schema"] = create_schema_from_function(
                    func.__name__,
                    func,
                    parse_docstring=func.__doc__,
                )
            return remote_func
        return func

    return wrapper



from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.runtime import Runtime


T = TypeVar("T", bound="ProviderConfig")


@_attrs_define
class ProviderConfig:
    """Model provider config

    Attributes:
        filename (Union[Unset, str]): The file name to use for the model
        presigned_url (Union[Unset, list[Any]]): The presigned URLs to upload the model to
        runtime (Union[Unset, Runtime]): Set of configurations for a deployment
    """

    filename: Union[Unset, str] = UNSET
    presigned_url: Union[Unset, list[Any]] = UNSET
    runtime: Union[Unset, "Runtime"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        filename = self.filename

        presigned_url: Union[Unset, list[Any]] = UNSET
        if not isinstance(self.presigned_url, Unset):
            presigned_url = self.presigned_url

        runtime: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.runtime, Unset):
            runtime = self.runtime.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if filename is not UNSET:
            field_dict["filename"] = filename
        if presigned_url is not UNSET:
            field_dict["presigned_url"] = presigned_url
        if runtime is not UNSET:
            field_dict["runtime"] = runtime

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.runtime import Runtime

        if not src_dict:
            return None
        d = src_dict.copy()
        filename = d.pop("filename", UNSET)

        presigned_url = cast(list[Any], d.pop("presigned_url", UNSET))

        _runtime = d.pop("runtime", UNSET)
        runtime: Union[Unset, Runtime]
        if isinstance(_runtime, Unset):
            runtime = UNSET
        else:
            runtime = Runtime.from_dict(_runtime)

        provider_config = cls(
            filename=filename,
            presigned_url=presigned_url,
            runtime=runtime,
        )

        provider_config.additional_properties = d
        return provider_config

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from beamlit.common.secrets import Secret

from . import kit


def github(name: str, *args):
    """This function kit is used to perform actions on Github."""
    github_token = Secret.get("GITHUB_TOKEN")
    if not github_token:
        raise ValueError("github_token missing from configuration.")

    modes = {}

    for func_name in dir(kit):
        if not func_name.startswith("_"):
            modes[func_name] = getattr(kit, func_name)
    if name not in modes:
        msg = f"Invalid mode: {name}"
        raise ValueError(msg)
    return modes[name](*args)
    return modes[name](*args)



# Import necessary modules
import ast
import asyncio
import functools
import importlib
import os
from logging import getLogger

from langchain_core.tools import Tool
from langgraph.checkpoint.memory import MemorySaver
from langgraph.prebuilt import create_react_agent

from beamlit.api.models import get_model
from beamlit.authentication import new_client
from beamlit.common.settings import init
from beamlit.errors import UnexpectedStatus
from beamlit.functions.mcp.mcp import MCPClient, MCPToolkit
from beamlit.functions.remote.remote import RemoteToolkit
from beamlit.models import Agent, AgentMetadata, AgentSpec

from .chain import ChainToolkit
from .chat import get_chat_model


def get_functions(dir="src/functions", from_decorator="function", remote_functions_empty=True):
    functions = []
    logger = getLogger(__name__)

    # Walk through all Python files in functions directory and subdirectories
    if not os.path.exists(dir):
        if remote_functions_empty:
            logger.warn(f"Functions directory {dir} not found")
        return []
    for root, _, files in os.walk(dir):
        for file in files:
            if file.endswith(".py"):
                file_path = os.path.join(root, file)
                # Read and compile the file content
                with open(file_path) as f:
                    try:
                        file_content = f.read()
                        # Parse the file content to find decorated functions
                        tree = ast.parse(file_content)

                        # Look for function definitions with decorators
                        for node in ast.walk(tree):
                            if (
                                not isinstance(node, ast.FunctionDef)
                                and not isinstance(node, ast.AsyncFunctionDef)
                            ) or len(node.decorator_list) == 0:
                                continue
                            decorator = node.decorator_list[0]

                            decorator_name = ""
                            if isinstance(decorator, ast.Call):
                                decorator_name = decorator.func.id
                            if isinstance(decorator, ast.Name):
                                decorator_name = decorator.id

                            if decorator_name == from_decorator:
                                # Get the function name and decorator name
                                func_name = node.name

                                # Import the module to get the actual function
                                spec = importlib.util.spec_from_file_location(func_name, file_path)
                                module = importlib.util.module_from_spec(spec)
                                spec.loader.exec_module(module)
                                # Check if kit=True in the decorator arguments
                                is_kit = False
                                if isinstance(decorator, ast.Call):
                                    for keyword in decorator.keywords:
                                        if keyword.arg == "kit" and isinstance(
                                            keyword.value, ast.Constant
                                        ):
                                            is_kit = keyword.value.value
                                if is_kit:
                                    kit_functions = get_functions(
                                        dir=os.path.join(root),
                                        from_decorator="kit",
                                        remote_functions_empty=remote_functions_empty,
                                    )
                                    functions.extend(kit_functions)

                                # Get the decorated function
                                if not is_kit and hasattr(module, func_name):
                                    func = getattr(module, func_name)
                                    if asyncio.iscoroutinefunction(func):
                                        functions.append(
                                            Tool(
                                                name=func.__name__,
                                                description=func.__doc__,
                                                func=func,
                                                coroutine=func,
                                            )
                                        )
                                    else:
                                        functions.append(
                                            Tool(
                                                name=func.__name__,
                                                description=func.__doc__,
                                                func=func,
                                            )
                                        )
                    except Exception as e:
                        logger.warning(f"Error processing {file_path}: {e!s}")
    return functions


def agent(
    agent: Agent | dict = None,
    override_chat_model=None,
    override_agent=None,
    mcp_hub=None,
    remote_functions=None,
):
    logger = getLogger(__name__)
    try:
        if agent is not None and not isinstance(agent, dict):
            raise Exception(
                'agent must be a dictionary, example: @agent(agent={"metadata": {"name": "my_agent"}})'
            )

        client = new_client()
        chat_model = override_chat_model or None
        settings = init()

        def wrapper(func):
            @functools.wraps(func)
            def wrapped(*args, **kwargs):
                return func(
                    settings.agent.agent,
                    settings.agent.chat_model,
                    settings.agent.functions,
                    *args,
                    **kwargs,
                )

            return wrapped

        # Initialize functions array to store decorated functions
        functions = get_functions(
            dir=settings.agent.functions_directory,
            remote_functions_empty=not remote_functions,
        )
        settings.agent.functions = functions

        if agent is not None:
            metadata = AgentMetadata(**agent.get("metadata", {}))
            spec = AgentSpec(**agent.get("spec", {}))
            agent = Agent(metadata=metadata, spec=spec)
            if agent.spec.model and chat_model is None:
                try:
                    response = get_model.sync_detailed(
                        agent.spec.model, environment=settings.environment, client=client
                    )
                    settings.agent.model = response.parsed
                except UnexpectedStatus as e:
                    if e.status_code == 404 and settings.environment != "production":
                        try:
                            response = get_model.sync_detailed(
                                agent.spec.model, environment="production", client=client
                            )
                            settings.agent.model = response.parsed
                        except UnexpectedStatus as e:
                            if e.status_code == 404:
                                raise ValueError(f"Model {agent.spec.model} not found")
                    else:
                        raise e
                except Exception as e:
                    raise e

                if settings.agent.model:
                    chat_model, provider, model = get_chat_model(agent.spec.model, settings.agent.model)
                    settings.agent.chat_model = chat_model
                    logger.info(f"Chat model configured, using: {provider}:{model}")

        if mcp_hub:
            for server in mcp_hub:
                try:
                    mcp_client = MCPClient(client, server)
                    toolkit = MCPToolkit(client=mcp_client)
                    toolkit.initialize()
                    functions.extend(toolkit.get_tools())
                except Exception as e:
                    logger.warn(f"Failed to initialize MCP server {server}: {e!s}")

        if remote_functions:
            for function in remote_functions:
                try:
                    toolkit = RemoteToolkit(client, function)
                    toolkit.initialize()
                    functions.extend(toolkit.get_tools())
                except Exception as e:
                    logger.warn(f"Failed to initialize remote function {function}: {e!s}")

        if agent.spec.agent_chain:
            toolkit = ChainToolkit(client, agent.spec.agent_chain)
            toolkit.initialize()
            functions.extend(toolkit.get_tools())

        if override_agent is None and len(functions) == 0:
            raise ValueError(
                "You must define at least one function, you can define this function in directory "
                f'"{settings.agent.functions_directory}". Here is a sample function you can use:\n\n'
                "from beamlit.functions import function\n\n"
                "@function()\n"
                "def hello_world(query: str):\n"
                "    return 'Hello, world!'\n"
            )

        if override_agent is None and chat_model is not None:
            memory = MemorySaver()
            agent = create_react_agent(chat_model, functions, checkpointer=memory)
            settings.agent.agent = agent
        else:
            settings.agent.agent = override_agent
        return wrapper
    except Exception as e:
        logger.error(f"Error in agent decorator: {e!s} at line {e.__traceback__.tb_lineno}")
        raise e



from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="StoreConfigurationOption")


@_attrs_define
class StoreConfigurationOption:
    """Store configuration options for a select type configuration

    Attributes:
        if_ (Union[Unset, str]): Conditional rendering for the configuration option, example: provider === 'openai'
        label (Union[Unset, str]): Store configuration option label
        value (Union[Unset, str]): Store configuration option value
    """

    if_: Union[Unset, str] = UNSET
    label: Union[Unset, str] = UNSET
    value: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        if_ = self.if_

        label = self.label

        value = self.value

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if if_ is not UNSET:
            field_dict["if"] = if_
        if label is not UNSET:
            field_dict["label"] = label
        if value is not UNSET:
            field_dict["value"] = value

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        if_ = d.pop("if", UNSET)

        label = d.pop("label", UNSET)

        value = d.pop("value", UNSET)

        store_configuration_option = cls(
            if_=if_,
            label=label,
            value=value,
        )

        store_configuration_option.additional_properties = d
        return store_configuration_option

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties




import ast

from beamlit.models import AgentChain, StoreFunctionParameter


def arg_to_list(arg: ast.List):
    value = []
    for v in arg.elts:
        value.append(format_value(v))
    return value

def format_value(v):
    if isinstance(v, ast.Constant):
        return v.value
    elif isinstance(v, ast.Dict):
        return arg_to_dict(v)
    elif isinstance(v, ast.List):
        return arg_to_list(v)

def arg_to_dict(arg: ast.keyword):
    value = {}
    for k, v in zip(arg.keys, arg.values):
        if isinstance(k, ast.Constant):
            value[k.value] = format_value(v)
    return value

def format_parameters(parameters: list[StoreFunctionParameter]) -> str:
    """
    Formats function parameters into YAML-compatible string.

    Args:
        parameters (list[StoreFunctionParameter]): List of parameter objects

    Returns:
        str: YAML-formatted string of parameters
    """
    if not parameters:
        return "[]"

    formatted = []
    for param in parameters:
        formatted.append(f"""
    - name: {param.name}
      type: {param.type_}
      required: {str(param.required).lower()}
      description: {param.description}""")

    return "\n".join(formatted)


def format_agent_chain(agentChain: list[AgentChain]) -> str:
    """
    Formats agent chain configuration into YAML-compatible string.

    Args:
        agentChain (list[AgentChain]): List of agent chain configurations

    Returns:
        str: YAML-formatted string of agent chain
    """
    if not agentChain:
        return "[]"
    formatted = []

    for agent in agentChain:
        formatted.append(f"""
      - agent: {agent.name}
        enabled: {agent.enabled}""")
        if agent.description:
            formatted.append(f"        description: {agent.description}")
    return "\n".join(formatted)


import asyncio
import warnings
from dataclasses import dataclass
from typing import Callable

import pydantic
import typing_extensions as t
from langchain_core.tools.base import BaseTool, ToolException

from beamlit.api.agents import list_agents
from beamlit.authentication.authentication import AuthenticatedClient
from beamlit.common.settings import get_settings
from beamlit.models import Agent, AgentChain
from beamlit.run import RunClient


class ChainTool(BaseTool):
    """
    Chain tool
    """

    client: RunClient
    handle_tool_error: bool | str | Callable[[ToolException], str] | None = True

    @t.override
    def _run(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        warnings.warn(
            "Invoke this tool asynchronousely using `ainvoke`. This method exists only to satisfy standard tests.",
            stacklevel=1,
        )
        return asyncio.run(self._arun(*args, **kwargs))

    @t.override
    async def _arun(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        settings = get_settings()
        result = self.client.run(
            "agent",
            self.name,
            settings.environment,
            "POST",
            json=kwargs,
        )
        return result.text

    @t.override
    @property
    def tool_call_schema(self) -> type[pydantic.BaseModel]:
        assert self.args_schema is not None  # noqa: S101
        return self.args_schema

class ChainInput(pydantic.BaseModel):
    inputs: str

@dataclass
class ChainToolkit:
    """
    Remote toolkit
    """
    client: AuthenticatedClient
    chain: list[AgentChain]
    _chain: list[Agent] | None = None

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    def initialize(self) -> None:
        """Initialize the session and retrieve tools list"""
        if self._chain is None:
            agents = list_agents.sync_detailed(
                client=self.client,
            ).parsed
            chain_enabled = [chain for chain in self.chain if chain.enabled]
            agents_chain = []
            for chain in chain_enabled:
                agent = [agent for agent in agents if agent.metadata.name == chain.name]
                if agent:
                    agent[0].spec.description = chain.description or agent[0].spec.description
                    agents_chain.append(agent[0])
            self._chain = agents_chain

    @t.override
    def get_tools(self) -> list[BaseTool]:
        if self._chain is None:
            raise RuntimeError("Must initialize the toolkit first")

        return [
            ChainTool(
                client=RunClient(self.client),
                name=agent.metadata.name,
                description=agent.spec.description or "",
                args_schema=ChainInput,
            )
            for agent in self._chain
        ]


from .error import HTTPError
from .logger import init as init_logger
from .secrets import Secret
from .settings import Settings, get_settings, init, init_agent
from .utils import copy_folder

__all__ = [
    "Secret",
    "Settings",
    "get_settings",
    "init_agent",
    "init",
    "copy_folder",
    "init_logger",
    "HTTPError",
]



from beamlit.api.models import get_model, list_models
from beamlit.authentication import get_authentication_headers, new_client
from beamlit.common.settings import init
from beamlit.models import Model
from beamlit.run import RunClient

settings = init()
client = new_client()
models = list_models.sync(client=client)
model: Model = get_model.sync(
    "gpt-3-5-turbo", client=client
)
# init_agent(client=client)
run_client = RunClient(client=client)
response = run_client.run(
    "model",
    "gpt-3-5-turbo",
    settings.environment,
    method="POST",
    path="/v1/chat/completions",
    json={
        "messages": [
            {
            "role": "user",
            "content": "Hello!"
            }
        ]
    }
)

if response.status_code == 200:
    print(response.json())
else:
    print(response.text)
print(get_authentication_headers(settings))
